// **********************************************************
// File:	MeshUtils.cpp - 15/9/2004
// Desc:	file description
//
// Copyright (c) 2004 by robal. All rights reserved.
// **********************************************************

#include "MeshUtils.h"
#include "Math3D.h"
#include <math.h>

// global variables declarations
std::vector<STriangleFix>	g_vectTriangles;

// functions declarations
void SetUpTriangleVectors( STriangleFix&, const std::vector<SVertexFix> );
void GetTriangleGradients( const SVertexFix, const SVertexFix, const SVertexFix,
							SVector3f&, SVector3f& );
void SmoothNormals( std::vector<SVertexFix>& );
void SmoothTangents(std::vector<SVertexFix>& );
void SmoothBinormals( std::vector<SVertexFix>& );
void FinalizeSmoothing( std::vector<SVertexFix>& );
void MakeOrthoTangent( std::vector<SVertexFix>& );

// functions definitions

// -- calculate triangle vectors --
void SetUpTriangleVectors( STriangleFix& triangle, const std::vector<SVertexFix> vectVertices )
{
	SVector3f	vectEdge1 = vectVertices[triangle.nIndices[1]].vectPos - vectVertices[triangle.nIndices[0]].vectPos;
	SVector3f	vectEdge2 = vectVertices[triangle.nIndices[2]].vectPos - vectVertices[triangle.nIndices[0]].vectPos;

	triangle.vectNormal = Vec3Cross( vectEdge1, vectEdge2 );
	Vec3Normalize( triangle.vectNormal );

	GetTriangleGradients( vectVertices[triangle.nIndices[0]],
							vectVertices[triangle.nIndices[1]],
							vectVertices[triangle.nIndices[2]],
							triangle.vectTangent, triangle.vectBinormal );
}

// -- calculate gradient vectors --
void GetTriangleGradients( const SVertexFix vert0, const SVertexFix vert1, const SVertexFix vert2,
							SVector3f& vectTangent, SVector3f& vectBinormal )
{
	SVector3f	vectP = vert1.vectPos - vert0.vectPos;
	SVector3f	vectQ = vert2.vectPos - vert0.vectPos;

	float	s1 = vert1.vectUV.x - vert0.vectUV.x;
	float	t1 = vert1.vectUV.y - vert0.vectUV.y;
	float	s2 = vert2.vectUV.x - vert0.vectUV.x;
	float	t2 = vert2.vectUV.y - vert0.vectUV.y;
	float	tmp = 0.0f;

	if( fabsf( s1 * t2 - s2 * t1 ) <= 0.0001f )
	{
		tmp = 1.0f;
	}
	else
	{
		tmp = 1.0f / ( s1 * t2 - s2 * t1 );
	}

	vectTangent.x = ( t2 * vectP.x - t1 * vectQ.x );
	vectTangent.y = ( t2 * vectP.y - t1 * vectQ.y );
	vectTangent.z = ( t2 * vectP.z - t1 * vectQ.z );

	vectTangent = vectTangent * tmp;

	vectBinormal.x = ( s1 * vectQ.x - s2 * vectP.x );
	vectBinormal.y = ( s1 * vectQ.y - s2 * vectP.y );
	vectBinormal.z = ( s1 * vectQ.z - s2 * vectP.z );

	vectBinormal = tmp * vectBinormal;
}

// -- smooth normals for each vertex --
void SmoothNormals( std::vector<SVertexFix>& vectVertices )
{
	int		nVertices = vectVertices.size();

	// for each vertex
	for( int i = 0; i < nVertices; i++ )
	{
		SVector3f	vectNormal( 0.0f, 0.0f, 0.0f );
		int			nFaces = vectVertices[i].vectTriangles.size();

		// for each face, which vertex belongs to
		for( int j = 0; j < nFaces; j++ )
		{
			int		nTriangle = vectVertices[i].vectTriangles[j];

			vectNormal = vectNormal + g_vectTriangles[nTriangle].vectNormal;
		}

		// after creating normal from all faces normalize it and store in final vertex
		Vec3Normalize( vectNormal );
		vectVertices[i].vectNormal = vectNormal;
	}
}

// -- smooth tangents for each vertex
void SmoothTangents( std::vector<SVertexFix>& vectVertices )
{
	int		nVertices = vectVertices.size();

	// for each vertex
	for( int i = 0; i < nVertices; i++ )
	{
		SVector3f	vectTangent( 0.0f, 0.0f, 0.0f );
		int			nFaces = vectVertices[i].vectTriangles.size();

		// for each face, which vertex belongs to
		for( int j = 0; j < nFaces; j++ )
		{
			int		nTriangle = vectVertices[i].vectTriangles[j];

			vectTangent = vectTangent + g_vectTriangles[nTriangle].vectTangent;
		}

		// after creating normal from all faces normalize it and store in final vertex
		Vec3Normalize( vectTangent );
		vectVertices[i].vectTangent = vectTangent;
	}
}

// -- smooth binormals for each vertex
void SmoothBinormals( std::vector<SVertexFix>& vectVertices )
{
	int		nVertices = vectVertices.size();

	// for each vertex
	for( int i = 0; i < nVertices; i++ )
	{
		SVector3f	vectBinormal( 0.0f, 0.0f, 0.0f );
		int			nFaces = vectVertices[i].vectTriangles.size();

		// for each face, which vertex belongs to
		for( int j = 0; j < nFaces; j++ )
		{
			int		nTriangle = vectVertices[i].vectTriangles[j];

			vectBinormal = vectBinormal + g_vectTriangles[nTriangle].vectBinormal;
		}

		// after creating normal from all faces normalize it and store in final vertex
		Vec3Normalize( vectBinormal );
		vectVertices[i].vectBinormal = vectBinormal;
	}
}

//-- smooth across mesh to find average vectors --
void FinalizeSmoothing( std::vector<SVertexFix>& vectVertices )
{
	int		nVertices = vectVertices.size();
	std::vector<SVector3f>		vectNormals;
	std::vector<SVector3f>		vectTangents;
	std::vector<SVector3f>		vectBinormals;

	for( int i = 0; i < nVertices; i++ )
	{
		SVector3f	vectTmpNormal( 0.0f, 0.0f, 0.0f );
		SVector3f	vectTmpTangent( 0.0f, 0.0f, 0.0f );
		SVector3f	vectTmpBinormal( 0.0f, 0.0f, 0.0f );

		for( int j = 0; j < nVertices; j++ )
		{
			// the same position - mean duplicated ?
			if( ( vectVertices[i].vectPos.x == vectVertices[j].vectPos.x ) &&
				( vectVertices[i].vectPos.y == vectVertices[j].vectPos.y ) &&
				( vectVertices[i].vectPos.z == vectVertices[j].vectPos.z ) )
			{
				vectTmpNormal = vectTmpNormal + vectVertices[j].vectNormal;
				vectTmpTangent = vectTmpTangent + vectVertices[j].vectTangent;
				vectTmpBinormal = vectTmpBinormal + vectVertices[j].vectBinormal;
			}
		}

		// normalize vectors
		Vec3Normalize( vectTmpNormal );
		Vec3Normalize( vectTmpTangent );
		Vec3Normalize( vectTmpBinormal );

		vectNormals.push_back( vectTmpNormal );
		vectTangents.push_back( vectTmpTangent );
		vectBinormals.push_back( vectTmpBinormal );
	}

	for( int i = 0; i < nVertices; i++ )
	{
		vectVertices[i].vectNormal = vectNormals[i];
		vectVertices[i].vectTangent = vectTangents[i];
		vectVertices[i].vectBinormal = vectBinormals[i];
	}
}

//-- make tangents as ortho axis --
void MakeOrthoTangent( std::vector<SVertexFix>& vectVertices )
{
	size_t	nVertices = vectVertices.size();
	size_t	i;

	for( i = 0; i < nVertices; i++ )
	{
		//we can use the Gram-Schmidt algorithm to find the newTangent and the newBinormal
		//newT = T - (N dot T)N
		//newB = B - (N dot B)N - (newT dot B)newT

		SVector3f	vTmpNormal = vectVertices[i].vectNormal;
		SVector3f	vTmpTangent = vectVertices[i].vectTangent;
		SVector3f	vTmpBinormal = vectVertices[i].vectBinormal;

		SVector3f	vNewTangent = vTmpTangent - Vec3Dot( vTmpNormal, vTmpTangent ) * vTmpNormal;
		SVector3f	vNewBinormal = vTmpBinormal - Vec3Dot( vTmpNormal, vTmpBinormal ) * vTmpNormal
												- Vec3Dot( vNewTangent, vTmpBinormal ) * vNewTangent;

		vectVertices[i].vectTangent = vNewTangent;
		vectVertices[i].vectBinormal = vNewBinormal;
		Vec3Normalize( vectVertices[i].vectTangent );
		Vec3Normalize( vectVertices[i].vectBinormal );

		// check vectors length and correct if need
		float	fTanLen = Vec3Length( vectVertices[i].vectTangent );
		float	fBinLen = Vec3Length( vectVertices[i].vectBinormal );

		if( ( fTanLen <= 0.001f ) ||
			( fBinLen <= 0.001f ) )
		{
			// tangent is sensible, so make binormal then
			if( fTanLen > 0.5f )
			{
				vectVertices[i].vectBinormal = Vec3Cross( vectVertices[i].vectNormal, vectVertices[i].vectTangent );
			}
			else
			{
				// binormal is sensible, so make tangent then
				if( fBinLen > 0.5f )
				{
					vectVertices[i].vectTangent = Vec3Cross( vectVertices[i].vectBinormal, vectVertices[i].vectNormal );
				}
				else
				{
					// none of them is sensible
					SVector3f	vectX( 1.0f, 0.0f, 0.0f );
					SVector3f	vectY( 0.0f, 1.0f, 0.0f );
					SVector3f	vectStart;

					float		fDotX = Vec3Dot( vectX, vectVertices[i].vectNormal );
					float		fDotY = Vec3Dot( vectY, vectVertices[i].vectNormal );

					if( fDotX < fDotY )
					{
						vectStart = vectX;
					}
					else
					{
						vectStart = vectY;
					}

					vectVertices[i].vectTangent = Vec3Cross( vectVertices[i].vectNormal, vectStart );
					vectVertices[i].vectBinormal = Vec3Cross( vectVertices[i].vectNormal, vectVertices[i].vectTangent );
				}
			}
		}
		else
		{
			if( Vec3Dot( vectVertices[i].vectBinormal, vectVertices[i].vectTangent ) > 0.999f )
			{
				vectVertices[i].vectBinormal = Vec3Cross( vectVertices[i].vectNormal, vectVertices[i].vectTangent );
			}
		}
	}
}

bool Dot3Mesh( std::vector<SVertexFix>& vectVertices, const std::vector<int> vectIndices )
{
	int		nTriangles;
	int		nVertices;

	// get sizes
	nVertices = vectVertices.size();
	nTriangles = vectIndices.size();

	// set up triangles
	for( int i = 0; i < nTriangles; i += 3 )
	{
		STriangleFix	sTriangle;

		// create new triangle
		sTriangle.nIndices[0] = vectIndices[i+0];
		sTriangle.nIndices[1] = vectIndices[i+1];
		sTriangle.nIndices[2] = vectIndices[i+2];

		// calculate tangent space vectors per triangle
		SetUpTriangleVectors( sTriangle, vectVertices );
		g_vectTriangles.push_back( sTriangle );

		// store triangle index, which vertex belongs to
		vectVertices[sTriangle.nIndices[0]].vectTriangles.push_back( g_vectTriangles.size() - 1 );
		vectVertices[sTriangle.nIndices[1]].vectTriangles.push_back( g_vectTriangles.size() - 1 );
		vectVertices[sTriangle.nIndices[2]].vectTriangles.push_back( g_vectTriangles.size() - 1 );
	}

	SmoothNormals( vectVertices );
	SmoothTangents( vectVertices );
	SmoothBinormals( vectVertices );
	FinalizeSmoothing( vectVertices );
	MakeOrthoTangent( vectVertices );
	
	return true;
}